import { Algebra, Foldable, Monad, Traversable } from './interfaces.js'
import { curry } from '../..//vendor/izuna/src/index.js'

/** @import { Applicative, ApplicativeTypeRef, Apply, Chain, Foldable as FoldableT, Functor, Monad as MonadT, Morphism, Traversable as TraversableT } from './types.js' */


/** @import { InferredMorphism } from './types.js' */

const kleisli = curry((f, g, x) => f(x).chain(g))

/**
 * @template T
 * @type {Pure<T> | Impure<T>} Free
 */
export class Free {
  /**
   * @template T
   * @param {T} value 
   */
  static of(value) {
    return new Pure(value)
  }
}

/** @template T */
export class Pure extends Algebra(Monad, Traversable) {
  #value

  /**
   * @param {T} value 
   */
  constructor(value) {
    super()
    this.#value = value
  }

  /**
   * @param {InferredMorphism<T>} f 
   */
  chain(f) {
    return f(this.#value)
  }

  /**
   * @param {InferredMorphism<T>} f 
   */
  map(f) {
    return pure(f(this.#value))
  }

  /**
   * @template U
   * @this {Pure<Morphism<U, T>>}
   * @param {Free<T>} b 
   */
  ap(b) {
    return b.map(this.#value)
  }

  /**
   * @type {FoldableT<T>['reduce']}
   */
  reduce(f, acc) {
    return f(acc, this.#value)
  }

  /**
   * @template U
   * @template {Applicative<U>} M
   * @template {ApplicativeTypeRef<U, M>} TypeRef
   * @param {TypeRef} _A
   * @param {(value: T) => M} f
   * @returns {Applicative<Free<U>>}
   */
  traverse(_A, f) {
    return f(this.#value).map(pure)
  }

  run() {
    return this.#value
  }
}

/**
 * @template T, [N=any]
 * @implements Functor<T>
 * @implements Chain<T>
 * @implements TraversableT<T>
 */
export class Impure extends Algebra(Monad, Foldable, Traversable) {
  #value
  #next

  /**
   * @param {T} value 
   * @param {(value: T) => Free<N>} f 
   */
  constructor(value, f) {
    super()
    this.#value = value
    this.#next = f
  }

  /**
   * @type {Chain<T>['chain']}
   * @template U
   * @param {(value: T) => Free<U>} f 
   * @returns {Chain<T>}
   */
  chain(f) {
    return impure(this.#value, kleisli(this.#next, f))
  }

  /**
   * @template {Functor<T>} U
   * @type {Functor<T>['map']}
   * @param {(value: T) => U} f 
   * @returns {Functor<T>}
   */
  map(f) {
    return impure(
      this.#value,
      y => f(y).map(f)
    )
  }

  /**
   * @type {Apply<T>['ap']}
   */
  ap(b) {
    return this.chain(f => b.map(f))
  }

  /**
   * @type {FoldableT<T>['reduce']}
   */
  reduce(f, acc) {
    const fb = f(acc, this.#value)
    const rest = this.#next(this.#value).reduce(f, acc)

    return rest.ap(
      fb.map(b => next => impure(b, () => next))
    )
  }

  /**
   * @type {TraversableT<T>['traverse']}
   * @template U
   * @template {Applicative<U>} M
   * @template {ApplicativeTypeRef<U, M>} TypeRef
   * @this {Impure<T, N>}
   * @param {TypeRef} A
   * @param {(value: T | N) => M} f
   * @returns {Applicative<Free<U>>}
   */
  traverse(A, f) {
    const fb = f(this.#value)
    const rest = this.#next(this.#value).traverse(A, f)

    return rest.ap(
      fb.map(b => next => impure(b, () => next))
    )
  }

  run() {
    return this.#next(this.#value).run()
  }
}

/**
 * @template T
 * @param {T} x
 * @returns {Pure<T>}
 */
const pure = x => new Pure(x)

/**
 * @template T
 * @param {T} x
 * @param {(value: T) => Free<any>} f
 * @returns {Impure<T>}
 */
const impure = (x, f) => new Impure(x, f)


/**
 * @template T
 * @param {T} x
 * @returns Impure<T>
 */
export const liftF = x => impure(x, pure)
import { id, Identity } from './identity.js'
import { Option, some } from './option.js'

const a = liftF(1).chain(() => liftF(2)).traverse(Option, x => some(1))
console.log(a.chain(x => x).run())

